package tr.edu.bilkent.cs.partitioning.ccp;

/**
 * PrefixSum interface.
 * This interface utilizes the methods that can be computed efficiently using a prefix summed array.
 *
 * Date: 20.04.2006
 */
public interface PrefixSum {
    /**
     * The number of tasks.
     * @return number of tasks.
     */
    int size();

    /**
     * The sum of weights up to item i, i.e., t_0 + t_1 + ... + t_i
     * @param i the last task for the summation.
     * @return the sum.
     */
    double get(int i);

    /**
     * Get a partial of the range (j, i], i.e.
     * \sum_{k=j+1}^i w_k.
     *
     * @param i the index of the last weight to be summed.
     * @param j the index of the task prior to the first task to be summed.
     * @return sum of the range.
     */
    double getSum(int i, int j);

    /**
     * sum of all items.
     *
     * @return total sum.
     */
    double getSum();

    /**
     * search for the prefix sum value.
     * returns the index of the nearest.
     *
     * @param v the prefix sum value to be searched.
     * @return nearest index.
     */
    int search(double v);

    /**
     * returns the index of the largest index that has a prefix sum not larger than v.
     *
     * @param v the prefix sum value to be searched.
     * @return the index of the largest index that has a prefix sum not larger than v.
     */
    int searchLessThan(double v);

    /**
     * returns the index of the smallest index that has a prefix sum not smaller than v.
     * @param v the prefix sum value to be searched.
     * @return the index of the smallest index that has a prefix sum not smaller than v.
     */
    int searchGreaterThan(double v);

    /**
     * returns the actual value of the item at p.
     * @param p the index of the desired item.
     * @return the actual value of the item at p.
     */
    double getAt(int p);

    /**
     * The underlying core prefix-summed array.
     * The implementations may return the internal structure for efficiency, thus, modifying this array
     * is highly discouraged.
     *
     * @return the internal prefix summed array.
     */
    double[] getArray();

    /**
     * The actual values.
     * The implementations of this method are expected to calculate the original array.
     * A new array is allocated and returned.
     *
     * @return the original array.
     */
    double[] getOriginalValues();
}
