package tr.edu.bilkent.cs.partitioning.ccp;

import java.io.*;
import java.util.Random;
import java.util.Arrays;

import tr.edu.bilkent.cs.partitioning.ccp.hetccp.cp.RunningStatisticsImpl;
import tr.edu.bilkent.cs.partitioning.ccp.hetccp.cp.RunningStatistics;

/**
 * Several utility and factory methods.
 *
 * Date: 22.09.2005
 */
@SuppressWarnings({"UtilityClass"})
public class Utils {
    /**
     * The default processor speed range. [1..PROCESSOR_RANGE]
     */
    private static final int PROCESSOR_RANGE = 20;

    /**
     * The empty ctor. Construction is disallowed for utility class. 
     */
    private Utils() {
    }

    /**
     * calculates the prefix sum of the array.
     * The array is cloned for prefix sum.
     *
     * @param array the core array.
     * @return the prefix sum of the array.
     */
    public static PrefixSum prefixSum(double[] array) {
        return prefixSum(array, true);
    }

    /**
     * calculates the prefix sum of the array.
     *
     * @param array the core array.
     * @param clone whether the array should be cloned, or the array can be used for the prefix sum.
     * @return the prefix sum of the array.
     */
    public static PrefixSum prefixSum(double[] array, boolean clone) {
        double[] result;
        if (clone)
            result = array.clone();
        else
            result = array;

        for (int i = 1; i < result.length; i++) {
            result[i] += result[i-1];
        }
        return new PrefixSumImpl(result);
    }

    /**
     * Reads the contents of the task array from a w-formatted file.
     * A w-formatted file is simply a text file: first line is an integer representing the number of tasks (n) and
     * the next n lines hold n doubles as the weights of the tasks.
     *
     * @param fileName the filename to be read.
     * @return the task array.
     * @throws IOException if some exception occurs while reading.
     */
    public static double[] readArray(String fileName) throws IOException {
        Closeable closeable = null;
        try {
            closeable = new FileInputStream(fileName);
            FileInputStream in = (FileInputStream) closeable;
            closeable = new InputStreamReader(in);
            InputStreamReader reader = (InputStreamReader) closeable;
            closeable = new BufferedReader(reader);
            BufferedReader br = (BufferedReader) closeable;
            String sline = br.readLine();
            String line = sline.trim();
            int n = Integer.parseInt(line);
            double[] a = new double[n];
            int nonzeroCount = 0;

            for (int i = 0; i < n; i++) {
                String dline = br.readLine();
                double v = Double.parseDouble(dline);
                if (Math.abs(v)>=Double.MIN_VALUE*10.0){
                    a[nonzeroCount]=v;
                    nonzeroCount++;
                }
            }
            double[] b = new double[nonzeroCount];
            System.arraycopy(a, 0, b, 0, nonzeroCount);
            br.close();
            return b;
        } finally {
            if (closeable!=null)
                closeable.close();
        }
    }

    /**
     * Returns a prefix summed processor speed array, with all processor speeds are 1.
     * @param p the number of processors.
     * @return the prefix summed processor speed array.
     */
    public static PrefixSum getUniformProcessors(int p) {
        double []pa = new double[p];
        for (int i=0; i<pa.length; i++)
            pa[i]=1;
        return prefixSum(pa);
    }

    /**
     * Returns random processors.
     * The chosen processor indices are uniformly distributed integers in the range [1, range].
     *
     * @param p the number of processors.
     * @param seed the seed.
     * @param range the range.
     * @return prefix summed processor speed array.
     */
    public static PrefixSum getRandomProcessors(int p, int seed, int range) {
        Random random = new Random(seed);
        double []pa = new double[p];
        for (int i=0; i<pa.length; i++)
            pa[i]=Math.abs(random.nextInt())%range+1;
        return prefixSum(pa);
    }

    /**
     * Returns random processors.
     * The chosen processor indices are uniformly distributed integers in the range [1, range].
     * Then the processors are sorted in increasing order.
     *
     * @param p the number of processors.
     * @param seed the seed.
     * @param range the range.
     * @return prefix summed processor speed array.
     */
    public static PrefixSum getRandomProcessorsOrdered(int p, int seed, int range) {
//        Random random = new Random(seed);
//        double []pa = new double[p];
//        for (int i=0; i<pa.length; i++)
//            pa[i]=Math.abs(random.nextInt())%range+1;
//        Arrays.sort(pa);
//        return prefixSum(pa);
        Random random = new Random(seed);
        int[] pa = new int[range];
        for (int i=0; i<p; i++){
            int x=Math.abs(random.nextInt())%range;
            pa[x]++;
        }
        int[] pp = new int[range];
        for (int i=0; i<pp.length; i++){
            pp[i]=i;
        }
        for (int i=0; i<range; i++){
            int x=Math.abs(random.nextInt())%range;
            int tmp = pp[i];
            pp[i] = pp[x];
            pp[x] = tmp;
        }
        double[] pr = new double[p];
        int next = 0;
        for (int i=0; i<range; i++){
            int e = pp[i];
            int count = pa[e];
            for (int j=0; j<count; j++){
                pr[next] = e+1;
                next++;
            }
        }
        if (next!=p)
            throw new IllegalStateException();
        return prefixSum(pr);
    }

    /**
     * Returns random processors.
     * The chosen processor indices are uniformly distributed doubles in the range [1, range].
     *
     * @param p the number of processors.
     * @param seed the seed.
     * @param range the range.
     * @return prefix summed processor speed array.
     */
    public static PrefixSum getRandomProcessorsDouble(int p, int seed, int range) {
        Random random = new Random(seed);
        double []pa = new double[p];
        for (int i=0; i<pa.length; i++)
            pa[i]=Math.abs(random.nextDouble()*(range-1)+1);
        return prefixSum(pa);
    }

    /**
     * Returns random processors.
     * The chosen processor indices are normally distributed doubles with mean and deviation as given.
     *
     * @param p the number of processors.
     * @param seed the seed.
     * @param mean the mean.
     * @param deviation the standard deviation.
     * @return prefix summed processor speed array.
     */
    public static PrefixSum getRandomProcessorsNormal(int p, int seed, double mean, double deviation) {
        Random random = new Random(seed);
        double []pa = new double[p];
        for (int i=0; i<pa.length; i++) {
            double a = random.nextGaussian();
            a = a * deviation +mean;
            pa[i]=Math.abs(a);
        }
        return prefixSum(pa);
    }

    /**
     * Returns random processors.
     * The chosen processor indices are normal-squared distributed doubles with mean and deviation as given.
     *
     * @param p the number of processors.
     * @param seed the seed.
     * @param mean the mean.
     * @param deviation the standard deviation.
     * @return prefix summed processor speed array.
     */
    public static PrefixSum getRandomProcessorsNormalSquare(int p, int seed, double mean, double deviation) {
        Random random = new Random(seed);
        double []pa = new double[p];
        int n = pa.length;
        int sqn= (int) Math.round(Math.sqrt(n));
        int sqm = n/sqn;
        for (int i=0; i<sqm; i++) {
            double a = random.nextGaussian();
            a = a * deviation +mean;

            for (int j=0; j<sqn; j++){
                double b = random.nextGaussian();
                b = b * deviation/sqm +a;
                pa[i*sqn+j] = Math.abs(b);
            }
        }
        return prefixSum(pa);
    }

    /**
     * performs a full garbage collection.
     */
    public static void fullGc(){
        Runtime runtime = Runtime.getRuntime();
        long last;
        long next = runtime.freeMemory();
        do{
            last = next;
            runtime.runFinalization();
            runtime.gc();
            next = runtime.freeMemory();
        }while (last>next);
    }

    /**
     * The default random processor chain factory.
     * @param p the number of processors.
     * @return the prefix summed processor-chain.
     */
    public static PrefixSum getRandomProcessors(int p) {
        return getRandomProcessors(p, 0, PROCESSOR_RANGE);
    }

    /**
     * returns a prefix-summed random ordering of the given prefix summed chain.
     *
     * @param origPSum original prefix sum.
     * @param seed seed.
     * @return the prefix summed reordered chain.
     */
    public static PrefixSum getRandomOrdering(PrefixSum origPSum, int seed) {
        double[] e = origPSum.getOriginalValues();
        Random random = new Random(seed);
        for (int i=0; i<e.length; i++){
            int next = random.nextInt(e.length);
            double tmp = e[i];
            e[i] = e[next];
            e[next] = tmp;
        }
        return prefixSum(e);
    }

    /**
     * max item of the prefix summed array.
     * Note that, this method searches the items one-by-one, resulting in an O(N)-time complexity.
     * @param pSum the prefix summed array.
     * @return max.
     */
    public static double getMax(PrefixSum pSum) {
        double max = 0;
        for (int i=0; i<pSum.size(); i++){
            double w = pSum.getAt(i);
            if (max<w)
                max = w;
        }
        return max;
    }

    /**
     * min item of the prefix summed array.
     * Note that, this method searches the items one-by-one, resulting in an O(N)-time complexity.
     * @param pSum the prefix summed array.
     * @return min.
     */
    public static double getMin(PrefixSum pSum) {
        double min = Double.MAX_VALUE;
        for (int i=0; i<pSum.size(); i++){
            double w = pSum.getAt(i);
            if (min>w)
                min = w;
        }
        return min;
    }

    /**
     * calculates and returns the statistics based on the sequence.
     * @param sequence the sequence to be investigated.
     * @return the statistics. 
     */
    public static RunningStatistics getStatistics(PrefixSum sequence){
        RunningStatistics statistics = new RunningStatisticsImpl();
        int n = sequence.size();
        for (int p=0;p<n; p++){
            double x = sequence.getAt(p);
            statistics.addObservation(x);
        }

        return statistics;
    }
}
