package tr.edu.bilkent.cs.partitioning.ccp.hetccp;

import tr.edu.bilkent.cs.ekt.pq.PriorityQ;
import tr.edu.bilkent.cs.ekt.pq.PriorityQImpl;
import tr.edu.bilkent.cs.partitioning.ccp.Partition;
import tr.edu.bilkent.cs.partitioning.ccp.PrefixSum;

import java.util.Comparator;

/**
 * The Bidding algorithm as described in:
 * A. Pinar, K. Tabak, C. Aykanat, One-Dimensional Partitioning for Heterogeneous Systems: Theory and Practice,
 * J. of Parallel and Distributed Computing, 2008.
 * <p>
 * Date: 10.03.2006
 */
public class Bidding implements HeterogeneousPartitionMethod {
    /**
     * calculates and returns partition.
     *
     * @param psArray    prefix summed task array
     * @param processors prefix summed processor array
     * @param minBid     minimum bid available.
     * @return the partition.
     */
    public static Partition calculatePartition(PrefixSum psArray, PrefixSum processors, double minBid) {
        int p = processors.size();
        int n = psArray.size();
        LRProbe lrProbe = Probe1DFactory.getLRProbeImpl();

        Partition partition;
        double lastBid;
        ProbeResult probeResult = lrProbe.probe(psArray, processors, minBid);
        partition = probeResult.getPartition();
        int[] seps = partition.getSeps();
        PriorityQ<BidStruct> pq = new PriorityQImpl<BidStruct>(new Comparator<BidStruct>() {
            public int compare(BidStruct o1, BidStruct o2) {
                return Double.compare(o1.value, o2.value);
            }
        });

        BidStruct[] bids = new BidStruct[p - 1];
        for (int i = 0; i < p - 1; i++) {
            double bid = getBid(psArray, processors, seps, i);
            bids[i] = new BidStruct(bid, i);
        }
        pq.buildHeap(bids);

        do {
            BidStruct struct = pq.peekMin();
            lastBid = psArray.getSum(partition.getSeparator(p - 1), partition.getSeparator(p - 2)) / processors.getAt(p - 1);
            if (minBid < lastBid) {
                for (int i = struct.getId(); i < p - 1; i++) {
                    double ep = processors.getAt(i);
                    int presep = (i == 0) ? -1 : seps[i - 1];
                    if (seps[i] < n - 1 && psArray.getSum(seps[i] + 1, presep) / ep <= minBid) {
                        seps[i] = psArray.searchLessThan(minBid * ep + psArray.get(presep));
                    }
                    while (seps[i] < n - 1 && psArray.getAt(seps[i] + 1) == 0.0) {
                        seps[i]++;
                    }
                    double bid = getBid(psArray, processors, seps, i);
                    pq.update(i, new BidStruct(bid, i));
                }
            }
        } while (minBid < lastBid);
        return partition;
    }

    /**
     * returns the bid of processor p, for given separators.
     *
     * @param psArray    prefix summed task array
     * @param processors prefix summed processor array
     * @param seps       the separators.
     * @param p          the processor index.
     * @return the bid.
     */
    public static double getBid(PrefixSum psArray, PrefixSum processors, int[] seps, int p) {
        double sum;
        if (p == 0)
            sum = psArray.get(seps[p] + 1);
        else if (seps[p] < psArray.size() - 1)
            sum = psArray.getSum(seps[p] + 1, seps[p - 1]);
        else
            sum = psArray.getSum(seps[p], seps[p - 1]);

        return sum / processors.getAt(p);
    }

    public Partition calculatePartition(PrefixSum psArray, PrefixSum processors, double wmax, double emin) {
        double minBid = psArray.getSum() / processors.getSum();
        return calculatePartition(psArray, processors, minBid);
    }

    public String getName() {
        return "Bidding  ";
    }

    private static class BidStruct {
        private int id;
        private double value;

        private BidStruct(double value, int id) {
            this.value = value;
            this.id = id;
        }

        public int getId() {
            return id;
        }
    }
}
