package tr.edu.bilkent.cs.partitioning.ccp.hetccp;

import tr.edu.bilkent.cs.partitioning.ccp.PrefixSum;

/**
 * Heterogeneous dynamic programming.
 * Date: 26.10.2005
 */
public class DP extends AbstractHeterogeneousPartitionMethod {

    public String getName() {
        return "DP";
    }

    /**
     * finds the optimal bottleneck value.
     *
     * @param psArray prefix-summed task array.
     * @param processors prefix-summed processor array.
     * @param wmax max of psArray.
     * @param emin min of processor array.
     * @return optimal bottleneck value.
     */
    @Override
    protected double findBStar(PrefixSum psArray, PrefixSum processors, double wmax, double emin) {
        // We need the actual values of processor speeds. Thus un-prefix the processors.
        double[] e = processors.getArray().clone();
        for (int i = e.length - 1; i > 0; i--) {
            e[i] -= e[i - 1];
        }

        double[] doubles = psArray.getArray();
        double[] previousW = doubles.clone();
        for (int i = 0; i < previousW.length; i++) {
            previousW[i] /= e[0];  // B[1,i]
        }

        double[] currentW = previousW.clone();

        int numberOfProcessors = e.length;  // P
        int numberOfTasks = previousW.length;   // N

        for (int p = 1; p < numberOfProcessors; p++) { // index starts from 0 in Java.
            double[] tmp = previousW;                  // shift of the DP array, to reduce memory consumption.
            previousW = currentW;                      // B[p-1]
            currentW = tmp;                            // B[p]
            int j = -1;
            for (int i = 0; i < numberOfTasks; i++) {
                if (psArray.getSum(i, j) / e[p] <= getBp(previousW, j)) {
                    currentW[i] = getBp(previousW, j);
                } else {
                    double sum;
                    do {
                        j++;
                        sum = psArray.getSum(i, j);
                    } while (sum / e[p] > getBp(previousW, j) && j < i);

                    double prebp = getBp(previousW, j);
                    double lastbp = sum / e[p];
                    if (lastbp == prebp) {
                        currentW[i] = prebp;
                    } else if (lastbp > prebp) {
                        throw new RuntimeException("not possible");
                    } else if (psArray.getSum(i, j - 1) / e[p] < prebp) {
                        j--;
                        currentW[i] = psArray.getSum(i, j) / e[p];
                    } else {
                        currentW[i] = prebp;
                    }
                }
            }
        }
        return currentW[numberOfTasks - 1];
    }

    /**
     * returns the bp[j] value.
     *
     * @param bp the current prefix-summed chain.
     * @param j the index.
     * @return bp[j] value.
     */
    private static double getBp(double[] bp, int j) {
        if (j == -1)
            return 0.0;
        return bp[j];
    }
}
