package tr.edu.bilkent.cs.partitioning.ccp.hetccp;

import tr.edu.bilkent.cs.partitioning.ccp.Partition;
import tr.edu.bilkent.cs.partitioning.ccp.PrefixSum;

/**
 * Exact Bisection algorithm as described in:
 * A. Pinar, K. Tabak, C. Aykanat, One-Dimensional Partitioning for Heterogeneous Systems: Theory and Practice,
 * J. of Parallel and Distributed Computing, 2008.
 *
 * Date: Jul 7, 2006
 */
public class ExactBisection extends AbstractHeterogeneousPartitionMethod {
    /**
     * The prober used in this implementation.
     */
    private static final LRProbe prober = Probe1DFactory.getLRProbeImpl();

    /**
     * name.
     */
    private static final String NAME = "E-BS      ";

    /**
     * calculates bound for the given task and processor chains.
     *
     * @param psArray    prefix-summed task chain.
     * @param processors prefix-summed processor array.
     * @return the bound.
     */
    public static Bound calculateBound(PrefixSum psArray, PrefixSum processors) {
        HeterogeneousPartitionMethod h3 = new MP();
        Partition partition = h3.calculatePartition(psArray, processors, 0, 0);
        double ub = partition.getPartitionValue();
        double lb = calculateInitialLb(psArray, processors);
        Bound bound = new BoundImpl(lb, ub);
        while (bound.getDiff() > 0.0) {
            double mid = bound.getMid();
            ProbeResult probeResult = prober.probe(psArray, processors, mid);
            if (probeResult.isProbed()) {
                Partition currentPartition = probeResult.getPartition();
                double partitionValue = currentPartition.getPartitionValue();
                bound.setUb(partitionValue);
            } else {
                double minBid = Double.MAX_VALUE;
                int numberOfProcessors = processors.size();
                for (int p = 0; p < numberOfProcessors - 1; p++) {
                    Partition currentPartition = probeResult.getPartition();
                    int[] separators = currentPartition.getSeps();
                    double bid = Bidding.getBid(psArray, processors, separators, p);
                    if (bid < minBid)
                        minBid = bid;
                }
                bound.setLb(minBid, true);
            }
        }
        return bound;
    }

    /**
     * calculates initial lower bound.
     *
     * @param psArray    prefix-summed task array.
     * @param processors prefix-summed processorarray.
     * @return initial lower bound.
     */
    private static double calculateInitialLb(PrefixSum psArray, PrefixSum processors) {
        return psArray.getSum() / processors.getSum();
    }

    @Override
    protected double findBStar(PrefixSum psArray, PrefixSum processors, double wmax, double emin) {
        Bound bound = calculateBound(psArray, processors);
        return bound.getUb();
    }

    /**
     * returns name.
     *
     * @return name.
     */
    public String getName() {
        return NAME;
    }

}
