package tr.edu.bilkent.cs.partitioning.ccp.hetccp;

import tr.edu.bilkent.cs.partitioning.ccp.Partition;
import tr.edu.bilkent.cs.partitioning.ccp.PrefixSum;

/**
 * Heterogeneous NicolPlus algorithm as described in:
 * A. Pinar, K. Tabak, C. Aykanat, One-Dimensional Partitioning for Heterogeneous Systems: Theory and Practice,
 * J. of Parallel and Distributed Computing, 2008.
 */
public class NicolPlus extends Nicol {
    private static final Probe1D PROBER = Probe1DFactory.getProbe1DImpl();

    @Override
    protected Crobber createCrobber(PrefixSum psArray, PrefixSum processors) {
        SeparatorBounds separatorBounds = SeparatorBoundsGenerator.getSeparatorBounds(psArray, processors);
        double ub = separatorBounds.getProbeValue();
        double totalSum = psArray.getSum();
        double lb = totalSum / processors.getSum();
        return new NicolPlusCrobber(new BoundImpl(lb, ub), separatorBounds);
    }

    @Override
    public String getName() {
        return "Nicol+";
    }

    private static class NicolPlusCrobber implements Crobber {
        private Bound bound;
        private SeparatorBounds separatorBounds;
        private LRProbe lrProbe = Probe1DFactory.getLRProbeImpl();

        NicolPlusCrobber(Bound bound, SeparatorBounds separatorBounds) {
            this.bound = bound;
            this.separatorBounds = separatorBounds;
        }

        public boolean cropInterval(PrefixSum psArray, PrefixSum processors, double candidateB, int mid, Interval interval, int p) {
            BoundCompareResult comp = bound.compare(candidateB);
            if (comp == BoundCompareResult.LESS) {
                interval.setFirst(mid + 1);
                return false;
            }
            if (comp == BoundCompareResult.UB || comp == BoundCompareResult.GREATER) {
                interval.setLast(mid);
                return true;
            }

            ProbeResult probeResult = lrProbe.probe(psArray, processors, candidateB);
            Partition partition = probeResult.getPartition();
            if (probeResult.isProbed()) {
                interval.setLast(mid);
                bound.setUb(candidateB);
                separatorBounds.updateWithProbedPartition(partition, p);
                return true;
            }
            // not probed.
            interval.setFirst(mid + 1);
            bound.setLb(candidateB, true);
            separatorBounds.updateWithFailedProbe(partition, p);
            return false;
        }

        public Interval getInterval(int first, int last, int p) {
            Interval interval = new IntervalImpl();
            interval.setFirst(Math.max(first, separatorBounds.getLow(p)));
            interval.setLast(Math.min(last, separatorBounds.getHigh(p)));
            return interval;
        }

        public double getNotProbedCandidate(PrefixSum psArray, PrefixSum processors, double candidate, int p) {
            BoundCompareResult comp = bound.compare(candidate);
            boolean probed;
            if (comp == BoundCompareResult.UB || comp == BoundCompareResult.GREATER) {
                probed = true;
            } else if (comp == BoundCompareResult.LESS) {
                probed = false;
            } else {
                ProbeResult probeResult = lrProbe.probe(psArray, processors, candidate);
                Partition partition = probeResult.getPartition();
                if (probeResult.isProbed()) {
                    bound.setUb(candidate);
                    separatorBounds.updateWithProbedPartition(partition, p);
                    probed = true;
                } else {
                    bound.setLb(candidate, true);
                    separatorBounds.updateWithFailedProbe(partition, p);
                    probed = false;
                }
            }
            if (probed)
                return candidate;
            return Double.MAX_VALUE;
        }
    }

}
