package tr.edu.bilkent.cs.partitioning.ccp.hetccp;

import tr.edu.bilkent.cs.partitioning.ccp.Partition;
import tr.edu.bilkent.cs.partitioning.ccp.PrefixSum;

/**
 * A utility class for separator bounds generator methods.
 */
public class SeparatorBoundsGenerator {
    /**
     * generates the initial bounds of separators for workload array on processors.
     * The algorithm is as described in Lemma 4.3 of :
     * A. Pinar, K. Tabak, C. Aykanat, One-Dimensional Partitioning for Heterogeneous Systems: Theory and Practice,
     * J. of Parallel and Distributed Computing, 2008.
     *
     * @param psArray    prefix-summed workload array
     * @param processors prefix-summed processor speeds.
     * @return initial separator bounds.
     */
    public static SeparatorBounds getSeparatorBounds(PrefixSum psArray, PrefixSum processors) {
        MP rb = new MP();
        Partition partition = rb.calculatePartition(psArray, processors, 0, 0);
        double ub = partition.getPartitionValue();
        LRProbe lrProbeImpl = Probe1DFactory.getLRProbeImpl();
        LRProbe rlProbeImpl = Probe1DFactory.getRLProbeImpl();

        int[] h1 = lrProbeImpl.probe(psArray, processors, ub).getPartition().getSeps();
        int[] l1 = rlProbeImpl.probe(psArray, processors, ub).getPartition().getSeps();

        double lb = calculateInitialLb(psArray, processors);

        int[] l2 = lrProbeImpl.probe(psArray, processors, lb).getPartition().getSeps();
        int[] h2 = rlProbeImpl.probe(psArray, processors, lb).getPartition().getSeps();

        int n = h1.length;
        for (int i = 0; i < n; i++)
            h1[i] = Math.min(h1[i], h2[i]);
        for (int i = 0; i < n; i++)
            l1[i] = Math.max(l1[i], l2[i]);
        return new SeparatorBoundsImpl(h1, l1, ub);
    }

    /**
     * returns initial lower bound for bottleneck value.
     * The algorithm is described in Section 4.2 of :
     * A. Pinar, K. Tabak, C. Aykanat, One-Dimensional Partitioning for Heterogeneous Systems: Theory and Practice,
     * J. of Parallel and Distributed Computing, 2008.
     *
     * @param psArray prefix-summed workload array
     * @param processors prefix-summed processor speeds.
     * @return initial lower bound on bottleneck
     */
    private static double calculateInitialLb(PrefixSum psArray, PrefixSum processors) {
        return psArray.getSum() / processors.getSum();
    }

}
